#!/usr/bin/env python

import sys,os,string
import wago

PDO_MAX = 5
case = 0
rpdo_parIdx = 0x1400
rpdo_mapIdx = 0x1600
tpdo_parIdx = 0x1800
tpdo_mapIdx = 0x1A00

def disable_all_mapping(nodeId):
        number = 1
        buffer = ""
        while number <= PDO_MAX:
                buffer += disable_tpdo(number, nodeId)
                buffer += disable_pdo_mapping("txPDO", number, tpdo_mapIdx)
                buffer += disable_pdo_mapping("rxPDO", number, rpdo_mapIdx)
                number += 1
        return buffer

def disable_pdo_mapping(type, pdo, idx):
        global case
        buffer = "\n\t\tcase %d:\n\t\t\t/* disable %s %d mapping */\n"%(case, type, pdo)
        buffer += "\t\t\tdata8 = 0x00;\n"
        buffer += "\t\t\twriteNetworkDictCallBack(d, nodeId, 0x%X, 0x00, 1, 0, &data8, checkSDO);"%(idx + pdo - 1)
        buffer += "\n\t\t\tbreak;"
        case += 1
        return buffer

def add_mapping(pdo, nr, width, src_subidx, src_idx, pdo_idx, pdo_subidx, type):
        global case
        buffer = "\n\t\tcase %d:\n\t\t\t/* map %s in %s %d */\n"%(case, type, pdo, nr)
        buffer += "\t\t\tmap[0] = 0x%.2X;\n\t\t\tmap[1] = 0x%.2X;\n\t\t\tmap[2] = 0x%s;\n\t\t\tmap[3] = 0x%s;\n"%(width,
                                                                                                                  src_subidx,
                                                                                                                  src_idx[4:6],
                                                                                                                  src_idx[2:4])
        buffer += "\t\t\twriteNetworkDictCallBack(d, nodeId, 0x%X, 0x%.2X, 4, 0, map, checkSDO);"%(pdo_idx + nr - 1,
                                                                                               pdo_subidx)
        buffer += "\n\t\t\tbreak;"
        case += 1
        return buffer

def disable_tpdo(pdo, nodeId):
        global case
        helper = 0x80000080 + (0x100 * pdo) + nodeId;
        buffer = "\n\t\tcase %d:\n\t\t\t/* disable txPDO %d */\n"%(case, pdo)
        buffer += "\t\t\tcobid = 0x%x;"%(helper)
        buffer += "\n\t\t\twriteNetworkDictCallBack(d, nodeId, 0x%X, 0x01, 4, 0, &cobid, checkSDO);"%(tpdo_parIdx + pdo - 1)
        buffer += "\n\t\t\tbreak;"
        case += 1
        return buffer

# set cobID of rpdos and tpdos and transmit types
def end_mapping(nodeId, rpdo_ids, tpdo_ids):
        buffer = ""
        tpdo = 0
        rpdo = 0
        while tpdo < len(tpdo_ids):
                buffer += set_cobID(tpdo_parIdx, "txPDO", tpdo + 1, tpdo_ids[tpdo])
                buffer += set_transmit_type(tpdo + 1, 1)
                tpdo += 1
        while rpdo < len(rpdo_ids):
                buffer += set_cobID(rpdo_parIdx, "rxPDO", rpdo + 1, rpdo_ids[rpdo])
                rpdo += 1
        return buffer

def set_transmit_type(nr, transmit_type):
        global case
        buffer = "\n\t\tcase %d:\n\t\t\t/* set txPDO %d transmit type to %d */\n"%(case, nr, transmit_type)
        buffer += "\t\t\tdata8 = %d;\n"%transmit_type
        buffer += "\t\t\twriteNetworkDictCallBack(d, nodeId, 0x%X, 0x02, 1, 0, &data8, checkSDO);"%(tpdo_parIdx + nr - 1)
        buffer += "\n\t\t\tbreak;"
        case += 1
        return buffer

def set_cobID(baseIdx, pdo, nr, id):
        global case
        buffer = "\n\t\tcase %d:\n\t\t\t/* set COB ID of %s %d */\n"%(case, pdo, nr)
        buffer += "\t\t\tcobid = 0x%X;"%id
        buffer += "\n\t\t\twriteNetworkDictCallBack(d, nodeId, 0x%X, 0x01, 4, 0, &cobid, checkSDO);"%(baseIdx + nr - 1)
        buffer += "\n\t\t\tbreak;"
        case += 1
        return buffer

def store_config():
        global case
        buffer = "\n\t\tcase %d:\n\t\t\t/* store config */"%case
        buffer += "\n\t\t\twriteNetworkDictCallBack(d, nodeId, 0x1010, 0x01, 4, 0, store, checkSDO);"
        buffer += "\n\t\t\tbreak;"
        case += 1
        return buffer

def set_operational(nodename):
        buffer = "\n\t\tcase %d:\n\t\t\t/* set operational */"%case
        buffer += "\n\t\t\tsetState(&%s_Data, Operational);"%nodename
        buffer += "\n\t\t\tmasterSendNMTstateChange(d, nodeId, NMT_Start_Node);"
        buffer += "\n\t\t\tconfig_step = 0;"
        buffer += "\n\t\t\tbreak;"
        return buffer

def get_check_func(nodename):
        buffer = "\nint config_node(CO_Data *d, UNS8 nodeId);"
        buffer += "\nvoid checkSDO(CO_Data* d, UNS8 nodeId)"
        buffer += "\n{\n\tUNS32 abortCode;\n\tUNS8 res;\n"
        buffer += "\n\tres = getWriteResultNetworkDict (d, nodeId, &abortCode);"
        buffer += "\n\tif(res != SDO_FINISHED)"
        buffer += '\n\t\tprintf("SDO abortCode: %x\\n",abortCode);\n'
        buffer += "\n\tcloseSDOtransfer(&%s_Data, nodeId, SDO_CLIENT);"%nodename
        buffer += "\n\tconfig_node(&%s_Data, nodeId);\n}"%nodename
        return buffer

def get_start_config_func(nodename):
        buffer = "\nint config_node(CO_Data *d, UNS8 nodeId)"
        buffer += "\n{\n\tUNS32 cobid = 0x00;"
        buffer += "\n\tUNS8 map[4] = {0x00, 0x00, 0x00, 0x00};"
        buffer += "\n\tUNS8 store[4] = {0x73, 0x61, 0x76, 0x65};"
        buffer += "\n\tUNS8 data8;\n"
        buffer += "\n\tswitch(config_step++){"
        return buffer

def get_end_config_func():
        buffer = "\n\t}"
        buffer += "\n\treturn 0;\n}"
        return buffer

# generate sdo transfers for pdo mapping
def get_mappings(pdomapping):
        mappings = ""
        rpdo_callbacks = ""
        indexes = wago.get_indexes()
        do_subidx = 0x01
        di_subidx = 0x01
        ai_subidx = 0x01
        ao_subidx = 0x01

        for type, nr, idx, pdo_subidx, name, index, width in pdomapping:
                if type == "rpdo":
                        # make slaves tpdo match masters rpdo
                        if name[0:2] == "DI":
                                mappings += add_mapping("txPDO", nr, width,
                                                        di_subidx, indexes["DI"],
                                                        tpdo_mapIdx, pdo_subidx, name)
                                di_subidx += 1
                        if name[0:2] == "AI":
                                mappings += add_mapping("txPDO", nr, width,
                                                        ai_subidx, indexes["AI"],
                                                        tpdo_mapIdx, pdo_subidx, name)
                                ai_subidx += 1
                if type == "tpdo":
                        # make slaves rpdo match masters tpdo
                        if name[0:2] == "DO":
                                mappings += add_mapping("rxPDO", nr, width,
                                                        do_subidx, indexes["DO"],
                                                        rpdo_mapIdx, pdo_subidx, name)
                                do_subidx += 1
                        if name[0:2] == "AO":
                                mappings += add_mapping("rxPDO", nr, width,
                                                        ao_subidx, indexes["AO"],
                                                        rpdo_mapIdx, pdo_subidx, name)
                                ao_subidx += 1
        return mappings

def get_config_functions(nodename, slavenodeid, tx_cobids, rx_cobids, pdomapping, pdomax):
        global PDO_MAX
        PDO_MAX = pdomax
        disable_mappings = disable_all_mapping(slavenodeid)
        config_functions = get_check_func(nodename) + get_start_config_func(nodename) + disable_mappings + \
                        get_mappings(pdomapping) + end_mapping(slavenodeid, tx_cobids, rx_cobids) + \
                        store_config() + set_operational(nodename) + get_end_config_func()

        return config_functions

