#!/usr/bin/env python

from node import *
from types import *

import gen_cfile
import confignode
import re, os

entry = {}
pdomapping = []

def get_includes(headerpath, file):
        buffer = ""
        if file == "actor":
                buffer = "#include <stdlib.h>"
                buffer += '\n#include "getopt.h"\n'
        buffer += '#include "%s"'%headerpath
        buffer += "\n#include <canfestival.h>"
        buffer += "\n#include <libpv.h>\n"
        return buffer

def get_global_vars(nodeid):
        buffer = "\nUNS8 mynodeid = 0x01;"
        buffer += "\nUNS8 slavenodeid = 0x%.2X;"%nodeid
        buffer += "\nUNS32 config_step = %d;"%confignode.case
        buffer += "\nint config_node(CO_Data *d, UNS8 nodeId);"
        return buffer

# adds a callback function for the pdo nr, vars contains all mapped variables
# for this pdo
def add_callback(type, nr, vars, comments):
        buffer = "\n\nUNS32 %s%d_Callback(CO_Data* d, const indextable * unused_indextable, UNS8 unused_Subindex)"%(type, nr)
        buffer += "\n{"
        buffer += "\n\tpv_lock(&pv_engine);\n"
        i = 0
        for var in vars:
                if comments[i] == "temp":
                        buffer += "\n\t*pv_%s = TOTEMP(%s);"%(var,var)
                elif comments[i] == "voltage":
                        buffer += "\n\t*pv_%s = TOVOL(%s);"%(var,var)
                else:
                        buffer += "\n\t*pv_%s = %s;"%(var,var)
                i += 1
        buffer += "\n\n\tpv_unlock(&pv_engine);"
        buffer += "\n\treturn 0;"
        buffer += "\n}"
        return buffer

def get_exit(entry):
        buffer = "\n\nvoid Exit(CO_Data* d, UNS32 id)"
        buffer += "\n{"
        buffer += "\n\tmasterSendNMTstateChange(&%(NodeName)s_Data, slavenodeid, NMT_Reset_Node);"%entry
        buffer += "\n\tsetState(&%(NodeName)s_Data, Stopped);"%entry
        buffer += "\n}\n"
        return buffer

def get_post_pdo(entry):
        buffer = "\n\nvoid %(NodeName)s_post_TPDO(CO_Data* d)"%entry
        buffer += "\n{"
        buffer += '\n\t/* printf("post_TPDO\\n"); */'
        buffer += "\n}"
        return buffer

def get_post_emcy(entry):
        buffer = "\n\nvoid %(NodeName)s_post_emcy(CO_Data* d, UNS8 nodeId, UNS16 errCode, UNS8 errReg)"%entry
        buffer += "\n{"
        buffer += "\n\tif (nodeId == slavenodeid && errCode == 0x0000){"
        buffer += "\n\t\tconfig_node(&%(NodeName)s_Data, nodeId);"%entry
        buffer += "\n\t}"
        buffer += "\n\telse"
        buffer += '\n\t\tprintf("Master received EMCY message. Node: %2.2x ErrorCode: %4.4x  ErrorRegister: %2.2x\\n", nodeId, errCode, errReg);'
        buffer += "\n}"
        return buffer

def get_post_slavebootup(entry):
        buffer ="\n\nvoid %(NodeName)s_post_slavebootup(CO_Data* d, UNS8 nodeId)"%entry
        buffer += "\n{"
        buffer += "\n\tconfig_node(&%(NodeName)s_Data, nodeId);"%entry
        buffer += "\n}"
        return buffer

def get_header(headerfilepath, entry, objdictpath, processVariables):
        entry["file_include_name"] = headerfilepath.replace(".", "_").upper()
        buffer = "#ifndef %(file_include_name)s"%entry
        buffer += "\n#define %(file_include_name)s"%entry
        buffer += "\n\n#include <libpv.h>"
        buffer += "\n#include <canfestival.h>"
        buffer += '\n#include "%s"\n\n'%objdictpath
        buffer += processVariables
        buffer += "pv_engine_t pv_engine;"
        buffer += "\n\nUNS32 config_step;"
        buffer += "\nint cf_init(int argc, char *argv[]);"
        buffer += "\nvoid cf_destroy(void);"
        buffer += "\nvoid cf_fire(void);"
        buffer += "\nvoid initNode(CO_Data * d, UNS32 id);"
        buffer += "\nvoid Exit(CO_Data* d, UNS32 id);"
        buffer += "\n\n#endif\n"
        return buffer

def start_init():
        buffer = "\n\nvoid initNode(CO_Data * d, UNS32 id)"
        buffer += "\n{"
        buffer += "\n\t/* canfestival callbacks */"
        buffer += "\n\t%(NodeName)s_Data.post_emcy = %(NodeName)s_post_emcy;"%entry
        buffer += "\n\t%(NodeName)s_Data.post_TPDO = %(NodeName)s_post_TPDO;"%entry
        buffer += "\n\t%(NodeName)s_Data.post_SlaveBootup = %(NodeName)s_post_slavebootup;"%entry
        buffer += "\n\n\t/* callbacks on pdo reception */"
        return buffer

def end_init():
        buffer = "\n\n\tsetNodeId(&%(NodeName)s_Data, mynodeid);"%entry
        buffer += "\n\tsetState(&%(NodeName)s_Data, Initialisation);"%entry
        buffer += "\n}"
        return buffer

# get entry attributes of index
def get_attributes(Node, index):
        global entry
        values = Node.GetEntry(index)
        subentry_entry = Node.GetSubentryInfos(index, 0)
        typename = Node.GetTypeName(subentry_entry["type"])
        typeentry = gen_cfile.GetValidTypeInfos(typename, [values])
        entry["callbacks"] = Node.HasEntryCallbacks(index)
        entry["subIndexType"] = typeentry[0]
        entry["name"] = gen_cfile.FormatName(subentry_entry["name"])
        entry["comment"] = Node.GetParamsEntry(index, None)["comment"]

# adds pdo mapping information of pdo nr to pdomapping
def get_pdo_mapping(Node, idx, type, nr):
        global pdomapping
        index = 0
        mask = 0x0000FFFF
        subIdxmax = Node.GetEntry(idx, 0x00)
        i = 1
        while i <= subIdxmax:
                index = Node.GetEntry(idx, i)
                name = Node.GetEntryName(index >> 16)
                pdomapping.append((type, nr, idx, i, name, index >> 16, index & mask))
                i += 1

def get_cf_init():
        buffer = "\n\nint cf_init(int argc, char *argv[])"
        buffer += "\n{\n\t/* canfestival init and start */"
        buffer += "\n\tint c;"
        buffer += '\n\ts_BOARD canopen_board = {"0", "125"};'
        buffer += '\n\textern UNS32 config_step;'
        buffer += '\n\tchar* LibraryPath="/usr/lib/libcanfestival_can_socket.so";\n'
        buffer += '\n\twhile ((c = getopt(argc, argv, "-h:b:r:l:c:v:d:s:8:j:")) != EOF)'
        buffer += "\n\t{\n\t\tswitch(c)\n\t\t{"
        buffer += "\n\t\t\tcase 'v':"
        buffer += "\n\t\t\tcase 'd':"
        buffer += "\n\t\t\tcase 's':"
        buffer += "\n\t\t\tcase '8':"
        buffer += "\n\t\t\tcase 'j':"
        buffer += "\n\t\t\t\tbreak;"
        buffer += "\n\t\t\tcase 'b':"
        buffer += "\n\t\t\t\tif (optarg[0] == 0){"
        buffer += "\n\t\t\t\t\tcf_usage(argv[0]);\n\t\t\t\t\treturn -1;"
        buffer += "\n\t\t\t\t}\n\t\t\t\tcanopen_board.busname = optarg;\n\t\t\t\tbreak;"
        buffer += "\n\t\t\tcase 'r':"
        buffer += "\n\t\t\t\tif (optarg[0] == 0){"
        buffer += "\n\t\t\t\t\tcf_usage(argv[0]);\n\t\t\t\t\treturn -1;"
        buffer += "\n\t\t\t\t}\n\t\t\t\tcanopen_board.baudrate = optarg;\n\t\t\t\tbreak;"
        buffer += "\n\t\t\tcase 'l':"
        buffer += "\n\t\t\t\tif (optarg[0] == 0){"
        buffer += "\n\t\t\t\t\tcf_usage(argv[0]);\n\t\t\t\t\treturn -1;"
        buffer += "\n\t\t\t\t}\n\t\t\t\tLibraryPath = optarg;\n\t\t\t\tbreak;"
        buffer += "\n\t\t\tcase 'c':"
        buffer += "\n\t\t\t\tif (optarg[0] == 0){"
        buffer += "\n\t\t\t\t\tcf_usage(argv[0]);\n\t\t\t\t\treturn -1;"
        buffer += "\n\t\t\t\t}\n\t\t\t\tif(atoi(optarg) == 1)\n\t\t\t\t\tconfig_step = 0;"
        buffer += "\n\t\t\t\tbreak;"
        buffer += "\n\t\t\tcase 'h':"
        buffer += "\n\t\t\tdefault:"
        buffer += "\n\t\t\t\tcf_usage(argv[0]);\n\t\t\t\treturn -1;\n\t\t}\n\t}"
        buffer += "\n\tif(!LoadCanDriver(LibraryPath)) {"
        buffer += '\n\t\tprintf("failed to load can driver\\n");'
        buffer += "\n\t\treturn -1;\n\t}\n"
        buffer += "\n\tif(!(canOpen(&canopen_board, &%(NodeName)s_Data))) {"%entry
        buffer += '\n\t\tprintf("failed to open can device\\n");'
        buffer += "\n\t\treturn -1;\n\t}\n"
        buffer += "\n\tTimerInit();"
        buffer += "\n\tStartTimerLoop(&initNode);"
        buffer += "\n\treturn 0;"
        buffer += "\n}"
        return buffer

def get_cf_usage():
        buffer = "\n\nvoid cf_usage(char *name)"
        buffer += "\n{"
        buffer += '\n\tfprintf(stderr, "\\nusage:\\t"'
        buffer += '\n\t"%s  [-v] -d <file> -s <file> [-j <port>] [-l <library>] [-b <bus name>] [-r <rate>]\\n\\n"'
        buffer += '\n\t"  -v, --verbose             Verbose mode of operation\\n"'
        buffer += '\n\t"  -d, --description <file>  Description XML File: specifies layout of the \\n"'
        buffer += '\n\t"                            process variable store file\\n"'
        buffer += '\n\t"  -s, --store <file>        Store file: this file holds the process variables\\n"'
        buffer += '\n\t"  -8, --8bit                use 8 bit backing store\\n"'
        buffer += '\n\t"  -j, --jvisu [<port>]      Start JVisu Socket Server, default 2202\\n"'
        buffer += '\n\t"\\n"'
        buffer += '\n\t"  -l,                       Can library, default /usr/lib/libcanfestival_can_socket.so\\n"'
        buffer += '\n\t"  -b,                       Bus name, default 0\\n"'
        buffer += '\n\t"  -r,                       Bit rate\\n"'
        buffer += '\n\t"  -c,                       device needs config, 1 == yes 0 == no (default)\\n"'
	buffer += '\n\t"\\n",'
        buffer += '\n\tname);'
        buffer += "\n}"
        return buffer

def get_cf_destroy():
        buffer = "\n\nvoid cf_destroy(void)"
        buffer += "\n{"
        buffer += "\n\tStopTimerLoop(&Exit);"
        buffer += "\n\tTimerCleanup();"
        buffer += "\n}"
        return buffer

def get_cf_fire():
        buffer = "\n\nvoid cf_fire(void)"
        buffer += "\n{"
        buffer += "\n\tif(getState(&%(NodeName)s_Data) == 0x05) {"%entry
        buffer += "\n\t\t%(NodeName)s_Data.PDO_status[0].last_message.cob_id = 0;"%entry
        buffer += "\n\t\tsendPDOevent(&%(NodeName)s_Data);"%entry
        buffer += "\n\t}"
        buffer += "\n}"
        return buffer

def genfile_content(objdictpath, Node, headerfilepath, pdomax):
        global entry
        global pdomapping
        # variables specified in "manufacturer specific" section for pdo mapping
        variablelist = [idx for idx in Node.GetIndexes() if 0x2000 <= idx <= 0xBFFF]

        entry["NodeName"] = Node.GetNodeName()
        # get slavenode id out of client sdo parameter
        slavenodeid = Node.GetEntry(0x1280, 3)

        # macros for the analogue terminals
        macros = "\n#define TOVOL(v)	((v) / 3276.8)"
        macros += "\n#define TOTEMP(v)	((v) / 10)\n"

        processVariables = ""
        callbackFunctions = ""
        registerCallbacks = ""

        tx_cobids = []
        txPDO_count = 0
        rx_cobids = []
        rxPDO_count = 0

        # get pdo infos - counts txpdos/rxpdos, gets cob ids out of pdo
        # parameter objects and mapping information from pdo mappings
        pdolist = [idx for idx in Node.GetIndexes() if 0x1400 <= idx <= 0x1BFF]
        for idx in pdolist:
                if 0x1400 <= idx <= 0x15FF:
                        # receive pdo parameter
                        rxPDO_count += 1
                        rx_cobids.append(Node.GetEntry(idx, 0x01))
                if 0x1600 <= idx <= 0x17FF:
                        # receive pdo mapping
                        get_pdo_mapping(Node, idx, "rpdo", idx - 0x1600 + 1)
                if 0x1800 <= idx <= 0x19FF:
                        # transmit pdo parameter
                        txPDO_count += 1
                        tx_cobids.append(Node.GetEntry(idx, 0x01))
                if 0x1A00 <= idx <= 0x1BFF:
                        # transmit pdo mapping
                        get_pdo_mapping(Node, idx, "tpdo", idx - 0x1A00 + 1)


        # this adds one callbackfunction per rpdo which gets registered at the
        # index of the last variable mapped in the rpdo
        callback_idx = 0
        i = 1
        while i <= rxPDO_count:
                variables = []
                comments = []
                for type, nr, idx, pdo_subidx, name, index, width in pdomapping:
                        get_attributes(Node, index)
                        if type == "rpdo" and nr == i:
                                variables.append(name)
                                comments.append(entry["comment"])
                                callback_idx = index
                # add callback function
                callbackFunctions += add_callback( "rpdo", i, variables, comments)
                # register callback function
                registerCallbacks += "\n\tRegisterSetODentryCallBack(&%s_Data, 0x%x, 0, &%s%d_Callback);"%(entry["NodeName"],callback_idx, "rpdo", i)
                i += 1

        # process variables for header file
        for index in variablelist:
                get_attributes(Node, index)

                if (entry["subIndexType"] == "UNS8"):
                        entry["subIndexType"] = "uint8_t*"
                elif (entry["subIndexType"] == "INTEGER16"):
                        entry["subIndexType"] = "float*"

                processVariables += "%(subIndexType)s pv_%(name)s;\n"%entry

        config_functions = confignode.get_config_functions(entry["NodeName"],
                                                           slavenodeid,
                                                           tx_cobids,
                                                           rx_cobids,
                                                           pdomapping,
                                                           pdomax)

        cfile = get_includes(headerfilepath, "config") + macros + get_global_vars(slavenodeid) + \
                        callbackFunctions + get_post_pdo(entry) + get_post_emcy(entry) + \
                        get_post_slavebootup(entry) + start_init() + registerCallbacks + \
                        end_init() + get_exit(entry) + config_functions + '\n'

        headerfile = get_header(headerfilepath, entry, objdictpath, processVariables)

        actor = get_includes(headerfilepath, "actor") + get_cf_usage() + get_cf_init() + \
                        get_cf_destroy() + get_cf_fire() + '\n'

        return cfile, headerfile, actor

# splits the basename and writes the generated files
def genfile(inpath, filepath, node, pdomax):
        try:
                headerfilepath = os.path.splitext(filepath)[0]+".h"
                cfilepath = os.path.splitext(filepath)[0]+"_config.c"
                actorfilepath = os.path.splitext(filepath)[0]+"_actor.c"
                objdictheaderfilepath = os.path.basename(inpath)
                content, header, actor = genfile_content(os.path.splitext(objdictheaderfilepath)[0]+".h",
                                                         node,
                                                         os.path.split(headerfilepath)[1],
                                                         pdomax)
                writefile(cfilepath, content)
                writefile(headerfilepath, header)
                writefile(actorfilepath, actor)
                return None
        except ValueError, message:
                return "Unable to Generate C File\n%s"%message

def writefile(filepath, content):
        cfile = open(filepath,"w")
        cfile.write(content)
        cfile.close()
